unit unzip;   {Unzips deflated, imploded and stored files}

{C code by info-zip group, translated to pascal by Christian Ghisler}
{based on unz51g.zip}
{Special thanks go to Mark Adler,
 who wrote the main inflate and explode code,
 and did NOT copyright it!!!}

interface

{$define assembler}      {use assembler!}

{$R-}         {No range checking}

{$ifdef windows}
uses wintypes,winprocs,strings,windos;
{$else}
uses strings,dos;
{$endif}

const   {Error codes returned by unzip}
  unzip_Ok=0;
  unzip_CRCErr=1;
  unzip_WriteErr=2;
  unzip_ReadErr=3;
  unzip_ZipFileErr=4;
  unzip_UserAbort=5;
  unzip_NotSupported=6;
  unzip_Encrypted=7;
  unzip_InUse=-1;

function GetSupportedMethods:longint;
{Checks which pack methods are supported by the dll} 
{bit 8=1 -> Format 8 supported, etc.}

function unzipfile(in_name:pchar;out_name:pchar;offset:longint;hFileAction:word;cm_index:integer):integer;
{usage:
 in_name:      name of zip file with full path
 out_name:     desired name for out file
 offset:       header position of desired file in zipfile
 hFileAction:  handle to dialog box showing advance of decompression (optional)
 cm_index:     notification code sent in a wm_command message to the dialog
               to update percent-bar
 Return value: one of the above unzip_xxx codes

 Example for handling the cm_index message in a progress dialog:

 unzipfile(......,cm_showpercent);

 ...

 procedure TFileActionDialog.wmcommand(var msg:tmessage);
 var ppercent:^word;
 begin
   TDialog.WMCommand(msg);
   if msg.wparam=cm_showpercent then begin
     ppercent:=pointer(lparam);
     if ppercent<>nil then begin
       if (ppercent^>=0) and (ppercent^<=100) then
         SetProgressBar(ppercent^);
       if UserPressedAbort then
         ppercent^:=$ffff
       else
         ppercent^:=0;
       end;
     end;
   end;
 end;
}


implementation

{*************************************************************************}

{$I z_global.pas}  {global constants, types and variables}
{$I z_tables.pas}  {Tables for bit masking, huffman codes and CRC checking}
{$I z_generl.pas}  {General functions used by both inflate and explode}
{$I z_huft.pas}    {Huffman tree generating and destroying}
{$I z_inflat.pas}  {Inflate deflated file}
{$I z_copyst.pas}  {Copy stored file}
{$I z_explod.pas}  {Explode imploded file}
{$I z_shrunk.pas}  {Unshrink function}

{***************************************************************************}

function GetSupportedMethods:longint;
begin
  GetSupportedMethods:=1+(1 shl 1)+(1 shl 6)+(1 shl 8);
  {stored, shrunk, imploded and deflated}
end;

{******************** main function: unzipfile *****************************}
{written and not copyrighted by Christian Ghisler}

function unzipfile(in_name:pchar;out_name:pchar;offset:longint;
  hFileAction:word;cm_index:integer):integer;
var e,r,i,err:integer;
    header:plocalheader;
    buf:array[0..80] of char;
    buf0:array[0..3] of char;
    timedate:longint;
    originalcrc:longint;    {crc from zip-header}
    ziptype,result:integer;
    p,p1:pchar;
    isadir:boolean;

begin
  {$ifdef windows}
  if inuse then begin
    {take care of crashed applications!}
    if (lastusedtime<>0) and
      (abs(gettickcount-lastusedtime)>30000) then begin {1/2 minute timeout!!!}
      {do not close files or free slide, they were already freed when application crashed!}
      inuse:=false;
      {memory for huffman trees is lost}
    end else begin
      unzipfile:=unzip_inuse;
      exit
    end;
  end;
  inuse:=true;
  {$endif}
  getmem(slide,wsize);     
  fillchar(slide[0],wsize,#0);
  assign(infile,in_name);
  filemode:=32;              {SHARE: others may read}
  {$I-}
  reset(infile,1);
  {$I+}
  if ioresult<>0 then begin
    freemem(slide,wsize);
    unzipfile:=unzip_ReadErr;
    inuse:=false;
    exit
  end;
  {$I-}
  seek(infile,offset);       {seek to header position}
  {$I+}
  if ioresult<>0 then begin
    freemem(slide,wsize);
    close(infile);
    unzipfile:=unzip_ZipFileErr;
    inuse:=false;
    exit
  end;
  header:=@inbuf;
  {$I-}
  blockread(infile,header^,sizeof(header^));  {read in local header}
  {$I+}
  if ioresult<>0 then begin
    freemem(slide,wsize);
    close(infile);
    unzipfile:=unzip_ZipFileErr;
    inuse:=false;
    exit
  end;

  if strlcomp(header^.signature,'PK'#3#4,4)<>0 then begin
    freemem(slide,wsize);
    close(infile);
    unzipfile:=unzip_ZipFileErr;
    inuse:=false;
    exit
  end;

  {calculate offset of data}
  offset:=offset+header^.filename_len+header^.extra_field_len+sizeof(tlocalheader);
  timedate:=header^.file_timedate;
  if (hufttype and 8)=0 then begin  {Size and crc at the beginning}
    compsize:=header^.compress_size;
    uncompsize:=header^.uncompress_size;
    originalcrc:=header^.crc_32;
  end else begin
    compsize:=maxlongint;           {Don't get a sudden zipeof!}
    uncompsize:=maxlongint;
    originalcrc:=0
  end;
  ziptype:=header^.zip_type;     {0=stored, 6=imploded, 8=deflated}
  if (1 shl ziptype) and GetSupportedMethods=0 then begin  {Not Supported!!!}
    freemem(slide,wsize);
    close(infile);
    unzipfile:=unzip_NotSupported;
    inuse:=false;
    exit;
  end;
  hufttype:=header^.bit_flag;
  if (hufttype and 1)<>0 then begin {encrypted}
    freemem(slide,wsize);
    close(infile);
    unzipfile:=unzip_Encrypted;
    inuse:=false;
    exit;
  end;

  reachedsize:=0;
  seek(infile,offset);

  assign(outfile,out_name);
  {$I-}
  rewrite(outfile,1);
  {$I+}
  err:=ioresult;
  {create directories not yet in path}
  isadir:=(out_name[strlen(out_name)-1]='/')
       or (out_name[strlen(out_name)-1]='\');
  if (err=3) or isadir then begin  {path not found}
    strcopy(buf,out_name);
    p1:=strrscan(buf,'\');
    if p1<>nil then inc(p1);  {pointer to filename}
    p:=strtok(buf,'\');
    if (p<>nil) and (p[1]=':') then begin
      strcopy(buf0,'c:\');    {set drive}
      buf0[0]:=p[0];
      {$ifdef windows}
      setcurdir(buf0);
      {$else}
      chdir(buf0);
      {$endif}
      p:=strtok(nil,'\');
    end;
    while (p<>nil) and (p<>p1) do begin
      {$ifdef windows}
      setcurdir(p);
      err:=doserror;
      {$else}
      {$I-}
      chdir(strpas(p));
      {$I+}
      err:=ioresult;
      {$endif}
      if err<>0 then begin
        {$ifdef windows}
        createdir(p);
        err:=doserror;
        {$else}
        {$I-}
        mkdir(strpas(p));
        {$I+}
        err:=ioresult;
        {$endif}
        if err=0 then
          {$ifdef windows}
          setcurdir(p);
          err:=doserror;
          {$else}
          {$I-}
          chdir(strpas(p));
          {$I+}
          err:=ioresult;
          {$endif}
      end;
      if err=0 then
        p:=strtok(nil,'\')
      else
        p:=nil;
    end;
    if isadir then begin
      freemem(slide,wsize);
      unzipfile:=unzip_Ok;    {A directory -> ok}
      close(infile);
      inuse:=false;
      exit;
    end;
    {$I-}
    rewrite(outfile,1);
    {$I+}
    err:=ioresult;
  end;

  if err<>0 then begin
    freemem(slide,wsize);
    unzipfile:=unzip_WriteErr;
    close(infile);
    inuse:=false;
    exit
  end;

  totalabort:=false;
  zipeof:=false;
  dlghandle:=hFileAction;
  dlgnotify:=cm_index;

  {$ifdef windows}
  messageloop;
  oldpercent:=0;
  {$endif}

  crc32val:=$FFFFFFFF;

  {Unzip correct type}
  case ziptype of
    0:result:=copystored;
    1:result:=unshrink;
    6:result:=explode;
    8:result:=inflate;
  else
    result:=unzip_NotSupported;
  end;
  unzipfile:=result;

  if (result=unzip_ok) and ((hufttype and 8)<>0) then begin {CRC at the end}
    dumpbits(k and 7);
    needbits(16);
    originalcrc:=b and $FFFF;
    dumpbits(16);
    needbits(16);
    originalcrc:=(b and $FFFF) shl 16;
    dumpbits(16);
  end;

  close(infile);
  close(outfile);
  crc32val:=not(crc32val);  {one's complement}
  if result<>0 then begin
    erase(outfile);
  end else if (originalcrc<>crc32val) then begin
    unzipfile:=unzip_CRCErr;
    erase(outfile);
  end else begin
    oldpercent:=100;       {100 percent}
    {$ifdef windows}
    if dlghandle<>0 then
      sendmessage(dlghandle,wm_command,dlgnotify,longint(@oldpercent));
    {$endif}
    filemode:=32;
    reset(outfile);
    setftime(outfile,timedate); {set zipped time and date of oufile}
    close(outfile);
  end;
  freemem(slide,wsize);
  inuse:=false;
end;

begin
  {$ifdef windows}
  inuse:=false;    {Not yet in use!}
  lastusedtime:=0; {Not yet used}
  {$endif}
  slide:=nil;      {unused}
end.
